/*
 * mm-naive.c - The fastest, least memory-efficient malloc package.
 * 
 * In this naive approach, a block is allocated by simply incrementing
 * the brk pointer.  Except for a size header (needed to make realloc
 * work), a block is pure payload. There are no footers.  Blocks are
 * never coalesced or reused. Realloc is implemented directly using
 * mm_malloc and mm_free.
 *
 * NOTE TO STUDENTS: Replace this header comment with your own header
 * comment that gives a high-level description of your solution.
 */
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <unistd.h>
#include <string.h>

#include "mm.h"
#include "memlib.h"

/*********************************************************
 * NOTE TO STUDENTS: Before you do anything else, please
 * provide your team information in the following struct.
 ********************************************************/
team_t team = {
    /* Team name */
    "",
    /* First member's full name */
    "",
    /* First member's email address */
    "",
    /* Second member's full name (leave blank if none) */
    "",
    /* Second member's email address (leave blank if none) */
    ""
};

/*
 * Choose the alignment, in bytes, that will be forced on all blocks
 * allocated.  All allocated blocks will be rounded up to a multiple
 * of this size; for that reason they will also automatically be
 * aligned on a boundary of this size.  (The latter follows only
 * because mem_sbrk returns aligend blocks.  On a real system, using
 * sbrk rather than mem_sbrk, blocks are aligned to an absurdly large
 * boundary, such as 4096 bytes.  However, in this lab the alignment
 * is guaranteed only to 8 bytes.)
 *
 * Normally, this should be set to the smallest value that will cause
 * all blocks to be aligned legally under the current compiler; i.e.,
 * 4 bytes on Linux, or 8 on Windows.
 */
#define ALIGNMENT 4

/*
 * The following pseudo-function generates an expression that rounds
 * up to the next highest multiple of ALIGNMENT.  However, it only
 * works if ALIGNMENT is a power of 2.  It is worth your time to
 * analyze how and why this expression works.
 *
 * If "size" is already aligned, the macro is a no-op.  This fact is
 * used in somme of the other functions.
 *
 * Incidentally, this #define illustrates a general principle of
 * writing macros (#defines) in C: every usage of any argument should
 * be parenthesized in case it is complicated, and the entire
 * expression should also be parenthesized in case it is used in a
 * larger statement.
 */
#define ALIGN(size) (((size) + ALIGNMENT - 1) & ~(ALIGNMENT - 1))

/*
 * The following typedef'd struct defines the header that is placed
 * before every block of allocated memory.  You may wish to expand
 * this header.  You can also use a similar declaration to define a
 * footer, if you choose.
 */
typedef struct {
    size_t size;		/* Size of this block, including header */
} header;

/*
 * The following macro generates an aligned (i.e., rounded-up) version
 * of the amount of space used for the header.  It will be used in
 * some subsequent macros.
 *
 * If you decide to create a footer, you will probably want to write
 * FOOTER_SIZE.  It may also be convenient to write OVERHEAD_SIZE as
 * the sum of HEADER_SIZE and FOOTER_SIZE (don't forget to
 * parenthesize the result!).
 */
#define HEADER_SIZE ALIGN(sizeof(header))

/*
 * The following macro converts a pointer to a block header into a
 * pointer to its payload.  To do so, it first typecasts the pointer
 * to "char *" (so that pointer arithmetic will be done on a byte
 * basis), and then casts it back to "void *" (because that's what
 * malloc will return).
 *
 * As with ALIGN, it is worth your time to study this macro so you can
 * understand what how it accomplishes its task.
 *
 * Even if you define a footer structure, there will be little need
 * for a FOOTER_TO_PAYLOAD macro.  However, you will probably want to
 * write HEADER_TO_FOOTER.
 */
#define HEADER_TO_PAYLOAD(p) ((void *)((char *)p + HEADER_SIZE))

/*
 * The following macro is the reverse of HEADER_TO_PAYLOAD.  It takes
 * a pointer to a payload and turns it into a correctly typecast
 * pointer to the corresponding header.
 *
 * Note that there probably won't be a need for PAYLOAD_TO_FOOTER, but
 * if you need it you can construct it fromm this macro and
 * HEADER_TO_FOOTER.
 */
#define PAYLOAD_TO_HEADER(p) ((header *)((char *)p - HEADER_SIZE))

/* 
 * mm_init - initialize the malloc package.
 */
int mm_init(void)
{
    return 0;
}

/* 
 * mm_malloc - Allocate a block by incrementing the brk pointer.
 *     Always allocate a block whose size is a multiple of the alignment.
 */
void *mm_malloc(size_t size)
{
    if (size == 0)
	return NULL;

    /*
     * The new size needs to allow for the header.
     */
    int newsize = ALIGN(size + HEADER_SIZE);
    header *hdr = (header *)mem_sbrk(newsize);
    if ((int)hdr == -1)
	return NULL;
    else {
	hdr->size = newsize;
        return HEADER_TO_PAYLOAD(hdr);
    }
}

/*
 * mm_free - Freeing a block does nothing in this allocator.
 */
void mm_free(void *ptr)
{
}

/*
 * mm_realloc - Implemented simply and fairly stupidly in terms of mm_malloc
 * and mm_free.
 */
void *mm_realloc(void *ptr, size_t newSize)
{
    header *oldhdr = PAYLOAD_TO_HEADER(ptr);
    void *newPayload;
    size_t copySize;
    
    /*
     * Handle corner cases in the specification.
     */
    if (newSize == 0) {
	if (ptr != NULL)
	    mm_free(ptr);
	return NULL;
    }
    else if (ptr == NULL) {
	return mm_malloc(newSize);
    }

    /*
     * Allocate new space, and copy it over.
     */
    newPayload = mm_malloc(newSize);
    if (newPayload == NULL)
      return NULL;
    copySize = oldhdr->size - HEADER_SIZE;
    if (newSize < copySize)
	copySize = newSize;
    memcpy(newPayload, ptr, copySize);
    mm_free(ptr);
    return newPayload;
}
